#include "mcmc.h"

long MCMC::samples = 0;
int MCMC::iteration = 0;
int MCMC::chain_run = 1;
long MCMC::sample_every = 0;
long MCMC::burn_in = 0;
bool MCMC::hill_climbing = 0;
bool MCMC::find_map = 0;
bool MCMC::model_averaging = 1;

MCMC::MCMC( int argc, char **argv)
      : Regression( argc, argv )
{
   initialise();
   for( int i = 1; i < argc; i++ )
   {
      if ( strstr( argv[i], "help" ) )
      {
         cout << "samples=" << samples << "   (The number of samples to be taken in any MCMC simulation)" << endl;
         cout << "every=" << sample_every << "   (The number of MCMC samples to be drawn for each one retained when thinning)" << endl;
         cout << "printoutevery=" << PrintOutEvery << "   (Print the detail after every this many MCMC samples)" << endl;
         cout << "numberofruns=" << number_of_runs << "   (The number of MCMC chains to be run)" << endl;
         cout << "hillclimbing=" << hill_climbing << "   (If non zero, accept an MCMC move only if it increases the log-posterior)" << endl;
         cout << "findmap=" << find_map << "   (If non zero, then report the maximum a posteriori model)" << endl;
         cout << "modelaveraging=" << model_averaging << "   (If zero, then report posterior modal statistics rather than means)" << endl;
         cout << "gibbssampling=" << gibbs_sampling << "   (Set non-zero if we wish to use gibbs sampler)" << endl;
         cout << "simulatedannealing=" << simulated_annealing << "   (Set non-zero if we wish to use simulated annealing)" << endl;
         cout << "simulatedannealingparameter=" << simulated_annealing_parameter << "   (Scale parameter for cooling schedule for simulated annealing)" << endl;
          cout << "simulatedannealingparameter2=" << simulated_annealing_parameter2 << "   (Shift parameter for cooling schedule for simulated annealing)" << endl;
         cout << "logposteriorfilename=" << Log_Posterior_Filename << "   (Name of file where log_posterior during MCMC sampling are written to)" << endl;
         return;
      }
      if( InitialStringMatch( argv[i], "samples=" ) )
         StringToLong( &samples, argv[i] );
      if( InitialStringMatch( argv[i], "every=" ) )
         StringToLong( &sample_every, argv[i] );
      if( InitialStringMatch( argv[i], "burnin=" ) )
         StringToLong( &burn_in, argv[i] );
      if( InitialStringMatch( argv[i], "printoutevery=" ) )
         StringToLong( &PrintOutEvery, argv[i] );
      if( InitialStringMatch( argv[i], "numberofruns=" ) )
         StringToLong( &number_of_runs, argv[i] );
      if( InitialStringMatch( argv[i], "hillclimbing=" ) )
         StringToBool( &hill_climbing, argv[i] );
      if( InitialStringMatch( argv[i], "findmap=" ) )
         StringToBool( &find_map, argv[i] );
      if( InitialStringMatch( argv[i], "modelaveraging=" ) )
         StringToBool( &model_averaging, argv[i] );
      if( InitialStringMatch( argv[i], "gibbssampling=" ) )
         StringToBool( &gibbs_sampling, argv[i] );
      if( InitialStringMatch( argv[i], "logposteriorfilename=" ) )
         StringToString( Log_Posterior_Filename, argv[i] );
      if( InitialStringMatch( argv[i], "simulatedannealing=" ) )
         StringToBool( &simulated_annealing, argv[i] );
      if( InitialStringMatch( argv[i], "simulatedannealingparameter=" ) )
         StringToDouble( &simulated_annealing_parameter, argv[i] );
      if( InitialStringMatch( argv[i], "simulatedannealingparameter2=" ) )
         StringToDouble( &simulated_annealing_parameter2, argv[i] );
   }
   if( samples <= burn_in )
      samples += burn_in;
   if( !PrintOutEvery )
      PrintOutEvery = ( samples - burn_in ) / 10;
   if( PrintOutEvery < 0 )
      PrintOutEvery = samples;
   mixing_print_out_every = sample_every;
   if( ( samples - burn_in ) / mixing_print_out_every > 10000 )
      mixing_print_out_every = ( samples - burn_in ) / 10000;
   if( simulated_annealing )
      find_map = true;
   if( hill_climbing || find_map )
      model_averaging = 0;
   if( !samples )
      simulated_annealing = false;
}

MCMC::MCMC( gsl_matrix* X1, gsl_matrix* Y1 )
      : Regression( X1, Y1 )
{
   initialise();
}

MCMC::MCMC( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1)
      : Regression( X1, Y1, D1 )
{
   initialise();
}


MCMC::~MCMC()
{
   zap_array( ave_predictions_mx, predictions_mx_rows, ave_predictions_mx_cols );
   zap_vector( proposal_cdf );
   zap_vector( TotalAcceptanceRB );
   zap_vector( TotalAcceptance );
   zap_vector( TotalIts );
}

void MCMC::initialise()
{
   accept = 0;
   force_accept = force_reject = false;
   posterior_cancels_with_proposal = false;
   last_iteration_accepted = 0;
   move_type = 0;
   number_of_moves = 0;
   proposal_cdf = new double[ 1 ];
   proposal_cdf[ 0 ] = 1;
   TotalAcceptanceRB = 0;
   TotalAcceptance = 0;
   TotalIts = 0;
   burn_in = 0;
   samples = 100000;
   sample_every = 1;
   number_of_runs = 1;
   ave_predictions_mx = 0;
   ave_predictions_mx_cols = 0;
   prediction_count = 0;
   predict_from_model = true;
   *Log_Posterior_Filename = 0;
   PrintOutEvery = 0;
   maximum_log_posterior = max_log_likelihood = maximum_run_posterior = -DBL_MAX;
   posterior_maximizing_iteration = 0;
   posterior_maximizing_run = 0;
   gibbs_sampling = 0;
   finished_mcmc_sampling = false;
   simulated_annealing = false;
   simulated_annealing_parameter = 10;
   simulated_annealing_parameter2 = 0;
   max_label_size = 0;
   average_log_posterior_ratio = 0;
   average_log_proposal_ratio = 0;
   average_proposed_log_posterior = 0;
   log_auxiliary_posterior = proposed_log_auxiliary_posterior = 0;
   average_acceptance_prob = 0;
   automatically_calculate_residuals = true;
}

void MCMC::sampler()
{
   if( !number_of_moves )
   {
      number_of_runs = 0;
      return;
   }
   run_ergodic_ave_normaliser = static_cast<double>( samples - burn_in )/ static_cast<double>( sample_every );
   ergodic_ave_normaliser = number_of_runs * run_ergodic_ave_normaliser;
   initialise_ergodic_averages();
   ofstream LogPosteriorStream( Log_Posterior_Filename, ios::out );
   ofstream ParameterStream( ParameterFilename, ios::out );
   if( simulated_annealing && print_to_screen )
   {
      cout << "Simulated annealing, cooling parameter(s) " << simulated_annealing_parameter;
      if( simulated_annealing_parameter2 > 0 )
         cout << ", " << simulated_annealing_parameter2 << endl;
      cout << endl;
   }
   finished_mcmc_sampling = false;
   maximum_run_posterior = log_posterior;
   for( chain_run = 1; chain_run <= number_of_runs; chain_run++ )
   {
      if( chain_run > 1 )
      {
         gsl_rng_set( rn, seed );
         srand( seed );
         reset_chain();
	 if( burn_in )
	 {
	    samples -= burn_in;
	    burn_in = 0;//burn in for first run of chain only
	 }
      }
      calculate_log_posterior();
      old_log_likelihood = log_likelihood = proposed_log_likelihood;
      old_log_prior = log_prior;
      old_log_posterior = log_posterior;
      if( chain_run == 1 )
         initial_log_posterior = log_posterior;
      if( log_posterior > maximum_log_posterior )
         maximum_log_posterior = log_posterior;
      previous_log_posterior = proposed_log_posterior = log_posterior;
      proposed_log_auxiliary_posterior = log_auxiliary_posterior;
      proposed_BIC = BIC;
      max_BIC = BIC;
      mcmc_sampling = true;
      iteration_time = time = clock();
      for( iteration = 1; iteration <= burn_in; iteration++ )
      {
         propose_a_move();
         if( !( LogAcceptanceRatio <= DBL_MAX ) )
            print_logacceptanceratio_warning();
         accept_or_reject_move();
         if( PrintOutEvery && iteration % PrintOutEvery == 0 )
	    print_out_every_output();
      }
      add_post_burn_in_moves();
      for( iteration = burn_in + 1; iteration <= samples; iteration++ )
      {
         propose_a_move();
         average_log_posterior_ratio += log_posterior_ratio;
         average_log_proposal_ratio += LogProposalRatio;
         average_proposed_log_posterior += proposed_log_posterior;
         if( force_accept )
            TotalAcceptanceRB[ move_type ]++;
         else if( !force_reject )
         {
            if( LogAcceptanceRatio < 0 )
            {
               if( LogAcceptanceRatio > -DBL_MAX )
               {
                  double acceptance_ratio = exp( LogAcceptanceRatio );
                  average_acceptance_prob += acceptance_ratio;
                  TotalAcceptanceRB[ move_type ] += acceptance_ratio;
               }
            }
            else if( LogAcceptanceRatio >= 0 )
            {
               average_acceptance_prob++;
               TotalAcceptanceRB[ move_type ]++;
            }
            else
               print_logacceptanceratio_warning();
         }
         TotalIts[ move_type ]++;
         accept_or_reject_move();
         if ( (iteration > burn_in) && !( ( iteration - burn_in) % sample_every ) )
         {
            update_ergodic_averages();
            base_update_ergodic_averages();
         }
         if( !( ( iteration - burn_in) % mixing_print_out_every ) )
         {
            if( strlen( Log_Posterior_Filename ) && LogPosteriorStream )
               LogPosteriorStream << log_posterior << endl;
            if( strlen( ParameterFilename ) && ParameterStream )
            {
               if( parameter_pointers.size() || integer_parameter_pointers.size() )
               {
                  for( list<double*>::iterator iter = parameter_pointers.begin(); iter != parameter_pointers.end(); iter++ )
                     ParameterStream << **iter << "\t";
                  for( list<int*>::iterator i_iter = integer_parameter_pointers.begin(); i_iter != integer_parameter_pointers.end(); i_iter++ )
                     ParameterStream << **i_iter << "\t";
                  ParameterStream << endl;
               }
            }
         }   
         if( PrintOutEvery && iteration % PrintOutEvery == 0 )
            print_out_every_output();
      }
      clock_t new_time = clock();
      double time_change = (double)( new_time - time ) / CLOCKS_PER_SEC;
      time = new_time;
      if( print_to_screen && samples )
         cout << "MCMC took " << time_change / 60 << " minutes." << endl;
      mcmc_sampling = false;
      end_of_chain_processing();
   }
   finished_mcmc_sampling = true;
   print_model_results();
   if( print_to_screen && samples )
      cout << "recalculated final log posterior : " << calculate_log_posterior() << endl;
}

double MCMC::calculate_log_prior()
{
   proposed_log_prior = 0;
   return proposed_log_prior;
}

void MCMC::propose_a_move()
{
   LogAcceptanceRatio = 0;
   LogProposalRatio = 0;
   accept = 0;
   force_accept = force_reject = false;
   posterior_cancels_with_proposal = false;
   draw_move_type();
   original_move_type = move_type;
   move();
   log_posterior_ratio = proposed_log_posterior - log_posterior;
   log_posterior_ratio += proposed_log_auxiliary_posterior - log_auxiliary_posterior;
   if( !posterior_cancels_with_proposal )
   {
      if( !simulated_annealing )
         LogAcceptanceRatio += log_posterior_ratio;
      else
         LogAcceptanceRatio += log_posterior_ratio / cooling_schedule();
   }
   if( !hill_climbing )
      LogAcceptanceRatio += LogProposalRatio;
//   LogAcceptanceRatio += proposed_log_auxiliary_posterior - log_auxiliary_posterior;
}

void MCMC::accept_or_reject_move()
{
   double log_u = 0;
   if( !hill_climbing & !(LogAcceptanceRatio >= 0) && !force_accept && !force_reject )
      log_u = log( gsl_ran_flat( rn, 0, 1 ) );
   
   if( force_accept || ( !force_reject && log_u <= LogAcceptanceRatio ) )
   {
      accept = 1;
      last_iteration_accepted = iteration;
      if ( iteration > burn_in )
         TotalAcceptance[ move_type ]++;
      execute_move();
      log_prior = proposed_log_prior;
      log_likelihood = proposed_log_likelihood;
      previous_log_posterior = log_posterior;
      log_posterior = proposed_log_posterior;
      log_auxiliary_posterior = proposed_log_auxiliary_posterior;
      BIC = proposed_BIC;
      if( log_posterior > maximum_log_posterior )
      {
         maximum_log_posterior = log_posterior;
         posterior_maximizing_iteration = iteration;
         posterior_maximizing_run = chain_run;
	 update_posterior_maximizing_model();
      }
   }
   else
   {
      undo_move();
      proposed_log_likelihood = log_likelihood;
      proposed_log_auxiliary_posterior = log_auxiliary_posterior;
      proposed_BIC = BIC;
   }
   post_move_updates();
}

void MCMC::print_acceptance_rates()
{
   int number_of_its = 0;
   int i;
   for( i = 0; i < number_of_moves; i++ )
      number_of_its += TotalIts[ i ];
   if( number_of_its <= 0 )
      return;
   cout << "Acceptance rates:" << endl;
   for( i = 0; i < number_of_moves; i++ )
   {
      if( TotalIts[ i ] > 0 )
      {
         cout << get_alligned_move_type_label( i ) << ":";
         cout << "   (RB) " << TotalAcceptanceRB[ i ] / (double)TotalIts[ i ] * 100 << "%";
         cout << "    (Actual) " << TotalAcceptance[ i ] << "/" << TotalIts[ i ] << endl;
      }
   }
   cout << endl;
   cout << "Average log ratio of proposed to current posterior     = " << average_log_posterior_ratio / (double)number_of_its << endl;
   cout << "Average value of proposed log posterior in MCMC        = " << average_proposed_log_posterior / (double)number_of_its << endl;
   cout << "Average log ratio of 'q' proposal probabilities in MH  = " << average_log_proposal_ratio / (double)number_of_its << endl;
   cout << "Average acceptance probability in MH                   = " << average_acceptance_prob / (double)number_of_its << endl;
   if( trans_dimensional_move_types.size() )
   {
      int num_td_its = 0;
      double td_acc = 0;
      for( list<int>::iterator iter = trans_dimensional_move_types.begin(); iter != trans_dimensional_move_types.end(); iter++ )
      {
         num_td_its += TotalIts[ *iter ];
         td_acc += TotalAcceptanceRB[ *iter ];
      }
      if( num_td_its )
         cout << "Average trans-dimensional acceptance probability in MH = " << td_acc / (double)num_td_its << endl;
   }
   cout << endl;
}

void MCMC::print_model_results()
{
   if( print_to_screen && samples )
   {
      cout << "\nMaximum log posterior (found after " << posterior_maximizing_iteration << " iterations";
      if( number_of_runs > 1 )
         cout << " of run " << posterior_maximizing_run;
      cout << ") = " << maximum_log_posterior << "      (Gain of " << maximum_log_posterior - initial_log_posterior << ")" << endl;
   }
   if( ave_predictions_mx )
      normalize_average_predictions_mx();
   write_predictions();
}

void MCMC::print_out_every_output()
{
   if( !print_to_screen )
      return;
   if( number_of_runs > 1 )
      cout << "Run " << chain_run << ", ";
   cout << "Iteration " << iteration  << endl;
   if( simulated_annealing )
      cout << "Temperature: " << cooling_schedule() << endl;
   if( PrintOutEvery == 1 )//if printing every iteration to screen, let's see the accpetance rates.
   {
      clock_t new_iteration_time = clock();
      double time_change = (double)( new_iteration_time - iteration_time ) * 1000.0 / CLOCKS_PER_SEC;
      iteration_time = new_iteration_time;
      cout << get_move_type_label() << "     (" << time_change << " milliseconds.)" << endl;
      cout << "Log(Acceptance prob.) = " << LogAcceptanceRatio << ",   ";
      if( accept )
         cout << "accepted\n(Previous Log Posterior = " << previous_log_posterior << ")" << endl;
      else
         cout << "rejected\n(Proposed Log Posterior = " << proposed_log_posterior << ")" <<endl;
   }
   cout << "Log Posterior = " << log_posterior << endl;
   cout << number_of_models_evaluated << " models evaluated." << endl;

//         if( iteration > burn_in )
   ModelPrintOut();
   print_acceptance_rates();
}

void MCMC::print_logacceptanceratio_warning()
{
   cout << "Warning: (Iteration " << iteration << ") Log acceptance ratio for ";
   cout << get_move_type_label() << " is " << LogAcceptanceRatio;
   cout << ", proposed posterior is " << proposed_log_posterior;
   cout << ", current posterior is " << log_posterior << endl;
   print_gsl_matrix( experiment_precisions );
   if( experiment_covariances )
      write_experiment_locations_to_file();
   exit(1);
}

void MCMC::write_predictions( ofstream *any_stream )
{
   bool prediction_averaging = !hill_climbing && !find_map && prediction_count && ave_predictions_mx;
   if( prediction_averaging )
      switch_values( predictions_mx, ave_predictions_mx );
   Regression::write_predictions( any_stream );
   if( automatically_calculate_residuals && strlen( Residuals_Filename ) )
   {
      if( !residuals )
         calculate_residuals();
      write_residuals();
   }
   if( prediction_averaging )
      switch_values( predictions_mx, ave_predictions_mx );
}
   
int MCMC::add_move()
{
   set_number_of_moves( number_of_moves + 1 );
   return number_of_moves - 1;
}

int MCMC::add_move( string new_label)
{
   move_type_labels[ number_of_moves ] = new_label;
   int label_size = new_label.size();
   if( label_size > max_label_size )
      max_label_size = label_size;
   return add_move();
}
void MCMC::add_number_of_moves( int add_number_of_moves )
{
   set_number_of_moves( number_of_moves + add_number_of_moves);
}
   
void MCMC::set_number_of_moves( int new_number_of_moves )
{
   number_of_moves = new_number_of_moves;
   number_of_moves_used = new_number_of_moves;   
   zap_vector( proposal_cdf );
   zap_vector( TotalAcceptanceRB );
   zap_vector( TotalAcceptance );
   zap_vector( TotalIts );
   proposal_cdf = new double[ number_of_moves ];
   TotalAcceptanceRB = new double[ number_of_moves ];
   TotalAcceptance = new int[ number_of_moves ];
   TotalIts = new int[ number_of_moves ];
   for(int i = 0; i < number_of_moves; i++ )
   {
      proposal_cdf[ i ] = (double)( i + 1 ) / (double)number_of_moves;
      TotalAcceptance[ i ] = 0;
      TotalAcceptanceRB[ i ] = 0;
      TotalIts[ i ] = 0;
   }
}

void MCMC::knock_out_move( int move_type_out )
{
   if( move_type_out == 0 )
      proposal_cdf[ move_type_out ] = 0;
   else
      proposal_cdf[ move_type_out ] = proposal_cdf[ move_type_out - 1 ];
   for(int i = 0; i < number_of_moves; i++ )
      proposal_cdf[ i ] = proposal_cdf[ i ] * (double) number_of_moves_used / (double)( number_of_moves_used - 1 );
   number_of_moves_used--;
}

void MCMC::draw_move_type()
{
   double u = gsl_ran_flat( rn, 0, 1 );
   move_type = 0;
   while( proposal_cdf[ move_type ] < u )
      move_type++;
}

void MCMC::move()
{
}

void MCMC::base_update_ergodic_averages()
{
   if( predict_from_model && ave_predictions_mx )
   {
      make_predictions();
      update_average_predictions_mx();
   }
   if( mean_experiment_covariances )
   {
      get_experiment_covariances();
      gsl_matrix_add( mean_experiment_covariances, experiment_covariances );
   }
}

void MCMC::update_average_predictions_mx()
{
   prediction_count++;
   for( int j = 0; j < predictions_update_cols; j++ )
   {
      int pred_col = predictions_mx_cols - 1 - j;
      int ave_pred_col =  ave_predictions_mx_cols - 1 - j;
      for( int i = 0; i < predictions_mx_rows; i++ )
         if( model_averaging )
            ave_predictions_mx[ i ][ ave_pred_col ] += predictions_mx[ i ][ pred_col ];
         else
            ave_predictions_mx[ i ][ ave_pred_col ] = predictions_mx[ i ][ pred_col ];
   }
}

void MCMC::normalize_average_predictions_mx()
{
   int i, j;
   for( i = 0; i < predictions_mx_rows; i++ )
   {
      if( prediction_count && model_averaging )
         for( j = 0; j < predictions_update_cols; j++ )
            ave_predictions_mx[ i ][ ave_predictions_mx_cols - 1 - j ] /= prediction_count;
      for( j = 0; j < ave_predictions_mx_cols - predictions_update_cols; j++ )
         ave_predictions_mx[ i ][ j ] = predictions_mx[ i ][ j ];
   }
}

string MCMC::get_move_type_label( int mt )
{
   string str;
   if( move_type_labels.find( mt ) != move_type_labels.end() )
      str = move_type_labels[ mt ];
   else
   {
      IntToString( &str, &mt );
      str = "Move Type " + str;
   }
   return str;
}

string MCMC::get_alligned_move_type_label( int mt )
{
   string str2 = get_move_type_label( mt );
   int label_size = str2.size();
   string str( max_label_size - label_size, ' ' );
   str += str2;
   return str;
}

double MCMC::cooling_schedule()
{
   return cooling_schedule( iteration, samples );
}

double MCMC::cooling_schedule( int it, int max_it )
{
   if( max_it == 1 )
      return 1;
   if( simulated_annealing_parameter2 > 0 )
      return exp( simulated_annealing_parameter2 - (double)it / max_it * simulated_annealing_parameter );
   return simulated_annealing_parameter / log( (double)( it + 1 ) );
}

bool MCMC::accept_since_iteration( int& reference_iteration )
{
   if( reference_iteration < last_iteration_accepted )
   {
      reference_iteration = iteration;
      return true;
   }
   return false;
}
