//reg.h

#if !defined( _reg_h )
#define _reg_h

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_sort.h>
#include <vector>
#include <algorithm>
#include "DataObj.h"
#define FILENAMELENGTH 100

#ifndef M_PI
#define M_PI 3.14159265358979323846264338327950288419716939937510
#endif
#define PI M_PI//3.14159265358979323846264338327950288419716939937510
#define LOG_PI log(M_PI)//1.14472988584940017414342735135305871164729481291531
#define LOG_2PI log(M_PI*2.0)//1.83787706640934548
#define LOG_SQRT_PI 0.5*log(M_PI)//0.5723649429247000870717135
using namespace std;


///
class Regression : public DataObj
{
  public:
   ///
   Regression( int argc, char **argv );
   ///
   Regression( gsl_matrix* X1, gsl_matrix* Y1 );// uses the matrix, matrix constructor of DataObj.h
   ///
   Regression( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 );// uses the matrix, matrix constructor of DataObj.h
   ///
   ~Regression();
   ///
   virtual Regression* create_new_regression( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 ) = 0;// overwritten at subsequent levels for creating a new regression of the type already initialised.

   void preliminary_regression_initialisation();
   
   void InitialiseParameters();

   void process_random_effect_variances( double rescale = 1 );

   void log_data();// changes Y to log( Y + 1 ).

   void logistic_transform_data();//changes Y to 1/(1+exp(-Y)).
   
   void difference_data( bool from_intercept = false );// changes X and Y to their first differences, or differences from the intercept if from_intercept=true.

   void append_test_points();//adds test_grid_size (squared if p_y>1)  many rows to X and Y, ranging Y between min and max values, used for prediction. ranges over grid of values on variables variable_to_predict_over and second_variable_to_predict_over.

   void normalise_target_data_blocks();//for each data block, calculate the norm (mean sum of squares) of it's targets.
   
   ///
   virtual void create_design_matrix();//here in the base class, default to pointing to Xptr.

   void create_subject_indices();//create subject_index int array holding the data row number where each subject begins.

   int get_subject_index( int sub_number );//returns the the index of the first row of the data matrices Yptr, Xptr relating to subject sub_number.

   void get_subject_row_range( int subject_num, int& start_row, int& end_row );//returns the data row range for that subject.

   gsl_matrix* get_targets( list<int> subject_list, int* exp_number = NULL );//returns a matrix containing the entries of *Yptr belonging to the subjects in subject_list and which have the same experiment numbers as *exp_number if given.

   double get_first_target( list<int> subject_list, int vbl = 0 );//returns the vbl'th variable of the first target associated to the list of subjects.
   
   gsl_matrix* get_target_averages( list<int> subject_list, int* exp_number = NULL );//returns a matrix containing the average entries of *Yptr belonging to the subjects in subject_list and which have the same experiment numbers as *exp_number if given.

   gsl_matrix* get_target_covariances( bool corr, bool residual_covariances );// calls get_target_covariances below for using all subjects in the data set.
   
   gsl_matrix* get_target_covariances( list<int> subject_list, bool corr = false, bool residual_covariances = false );//finds the covariances (or correlations if corr=true) of the targets (treating the data_block_sizes_sum units as columns) for the list of subjects, discarding missing values.
   
   void calculate_design_matrix_column_sums();
   
   void create_experiment_numbers_vector();//create gsl::vector holding the experiment number of each data_block.
   
   void find_experiment_sds();//find the standard deviations of the targets for each experiment.

   void standardize_targets_by_experiment();//divide the targets by the standard deviation for the corresponding experiment.
   
   int get_experiment_number( int obsn_num );//returns the experiment from which the observation obsn_num came from.

   double get_experiment_distance( int e1, int e2 );//returns the Euclidean distance between the locations of experiments e1 and e2.

   double get_experiment_dot_product( int e1, int e2 );//returns the dot product of the experiment locations for the two experiments.

   virtual double get_experiment_correlation( int e1, int e2 );//returns the correlation between the two experiment numbers induced by their locations.

   void get_experiment_correlations();//works out the values of  a matrix of correlations between the experiments using the function above, and puts them in the gsl_matrix experiment_covariances.
   
   virtual double get_experiment_covariance( int e1, int e2 );//returns the covariance between the two experiment numbers induced by their locations.

   void get_experiment_covariances();//works out the values of experiment_convariances, a matrix of covariances between the experiments, using the function above.
   void get_experiment_precisions();//works out the values of experiment_precisions, the inverse of the  matrix of covariances between the experiments.

   void update_experiment_precisions( int e1 );// updates the experimental precision (and then covariance) matrix after a change to experimental location e1.

   double get_experiment_location_norm( int e1 );//returns squared Eculidean distance from origin of location of experiment e1.

   double get_experiment_location_squared_distance_from_mean( int e1 );//returns squared Eculidean distance of location of experiment e1 from its mean.
   
   void reverse_experiment_sign( int e1 );//changes the sign of experiment e1 - if e1 = 0 though we change the sign of all the other experiments instead, so experiment_signs[ 0 ] stays fixed at 1.

   gsl_matrix* calculate_correlation_by_experiment( gsl_matrix* mx );//finds the correlation between experiments for data mx, with ith row of mx belonging to experiment experiment_num[ i ]. Assumes each experiment EQUALLY REPRESENTED (i.e. same number of obsns) in experiment_num/mx.

   void experiment_covariances_to_distances( gsl_matrix* mx );//change the correlations in mx into the corresponding distances.

   void experiment_covariances_to_locations( gsl_matrix* mx );// calls one of the two functions below.
   
   void experiment_mds_correlations_to_locations( gsl_matrix* mx );//convert the correlations to distances, then perform multidimensional scaling to recover locations.

   virtual void reset_experiment_location_parameters();//sets experiment location parameters equal to the average of the associated locations.
   
   void experiment_cholesky_correlations_to_locations( gsl_matrix* mx );//convert correlation matrix to locations using Cholesky decomposition.

   void sample_locations();//samples the experimental locations from their distribution.
   
   void find_experiment_location_sums();//adds the experiment locations together (for each group in the partition of experiments, if there is one).

   void delete_missing_points();// deletes the points from the matrices according to missing_values.

   void bootstrap_data();//replace the data with replace_with_bootstrap_samples many samples (with replacement) from the data.

   void thin_the_data();//only keep those data points for each subject whose row number is a multiple of thin_data (including the zeroth row)

   void switch_subject_and_expt_structures();//revert data_block_sizes to data_block_sizes_original, etc, or vice-versa.
   
   int get_number_of_design_variables();//returns the number of columns in the design mx D.
   ///
   int get_number_of_data_blocks(){ return num_data_blocks; }

   int get_number_of_subjects(){ return num_subjects; }   
///
   virtual double calculate_log_likelihood() = 0;
   ///
   virtual double calculate_log_prior();
   ///
   double calculate_log_posterior();

   double* get_x_min();

   double* get_x_max();

   double* get_x_bar();

   double* get_y_bar();

   virtual double function_evaluate( double* v, int dim_v = 1 );//evaluates the regression function at v.

   virtual double integral_function_evaluate( double* v, int integral_order = 0, int integral_component = 0 );//returns the integral_order'th integral of the regression function in the intergral_component direction evaluated at v.

   virtual void create_design_point( double* v ){ design_point = v;}

   virtual double function_variance_evaluate( double* v );//calculates the variance of the regression function at v ( v is assumed to be a vector of lenght p ).

   void find_unique_data_points();// find all the unique entries in the input matrix.

   void create_test_grid( int col1 = 0, int col2 = -1 );//make grid of test points for make_grid_predictions(...).
   
   void make_grid_predictions( int col = 0, int ycol = 0 );//updates predictions matrix as the col'th (ycol'th) variable in X (Y) ranges from 0 to 1

   virtual void write_predictions( ofstream *any_stream = NULL );//creates the predictions_mx and writes it to any_stream or by default to PredictionFilename.   

   void write_residuals( ofstream *any_stream = NULL );//writes the residuals matrix to any_stream.
      
   void write_experiment_locations_to_file( bool principal_components = true );//if principal_components true, write the principal components of the experiment locations to file, otherwise write the locations themselves

   virtual void make_predictions(){;}//virtual function used by write_predictions, to be overwritten.

   void create_residuals_matrix();//intialises residuals matrix.
   
   void calculate_residuals();// calculates the residuals matrix for the fitted function.

   void calculate_obsn_residual( int obsn );// calcualtes the residual for observation obsn.
   
   gsl_matrix* get_residuals();

   double get_r_squared();
   
   void calculate_correlation_matrix(); // calculates estimated correlation matrix and writes to file.
  
   void calculate_residual_variogram();//calculates variogram of residuals and writes the mean function as two columns, x and y, to file Variogram_Filename.
   
   virtual void output_residual_variogram( gsl_matrix* x_diffs, gsl_matrix* residual_diffs );// writes the variogram values to Variogram_Filename, overwritten at higher level to fit a curve and writing that to file.
   
   void create_gauss_laguerre_matrices();//intialises w_gl and x_gl.
   
   virtual double gauss_laguerre_integrand( int i ){ return gsl_matrix_get( x_gl, i, 0 ); }//default function for gauss-laguerre integration
   
   double gauss_laguerre_integrate( double (*integrand)( double ) );//does the gauss-laguerre integeration.
   
   double gauss_laguerre_integrate( double (*integrand)( int ) );//does the gauss-laguerre integration. Assumes the integrand is a class member and will have access to x_gl, so only takes the index as an argument.

   double gauss_laguerre_integrate();
   
   bool with_probability( double d );//returns true with probability min(d,1).

   virtual void set_active_dimension( int activ_dim );

   void count_subject_experiment_missing_values();//find the number of data points missing from each expt for each subject.

   void get_pairwise_sample_variances();//for each pair of subjects, calculate the sample error variance when they form a two-subject cluster.

 private:
   ///
   char TestInputFilename[FILENAMELENGTH];
   ///
   char TestTargetFilename[FILENAMELENGTH];
   ///
   char PrecisionFilename[FILENAMELENGTH];
   ///
   char InitialParameterFilename[FILENAMELENGTH];
   ///
   char Correl_Filename[FILENAMELENGTH];// file for writing the correlation matrix to.
   ///

 protected:
   ///
   char PredictionFilename[FILENAMELENGTH];

   char Subject_Predictions_Filename[FILENAMELENGTH];

   ///
   int regression_n;//the number of data rows in the regression (n, unless other test points have been added).

   gsl_rng* rn; // GSL random number generator.

   const gsl_rng_type* rn_type;// GSL random number generator type.
   
   gsl_matrix* D; //pointer to Design matrix, may or may not be equal to Xptr.

   gsl_matrix* D_column_sums;//pointer to a matrix holding the column sums of D.
   
   gsl_matrix* residuals;// pointer to residuals of *Yptr to the fitted function.

   gsl_matrix* beta; //pointer to matrix of parameters

   list< double* > parameter_pointers;//pointers to model parameters.
   list< int* > integer_parameter_pointers;
   
   gsl_matrix *w_gl, *x_gl;//matrices to hold weights and abscissae for gauss laguerre estimation.

   double gauss_laguerre_multiplier;//multiplier for x_gl, divisor for w_gl.

   int num_design_cols;//number of columns in the design matrix D.
   
   int dim_gl;//size of Gauss-Laguerre arrays.

   int design_point_dim;//dimension of a design point - p by default.

   double* design_point;//a spare single design point, used for making predictions.
   
   double log_likelihood;

   double log_likelihood_constant;//a constant contribution to the log_likelihood.

   double log_prior;

   double log_posterior;

   double BIC;

   double r_squared;

   static unsigned long number_of_models_evaluated;//static variable keeping track of the number of times the likelhood function of a model has been calculated.

   static int number_of_regressions;// static variable keeping track of the number of regression objects initialised.

   static long seed;

   bool mcmc_sampling;//zero unless mcmc sampling has been initiated.

   ///
   double** predictions_mx;// matrix to hold predictions to be written to PredictionFilename.

   double** test_grid_points;//matrix holding points at which predictions are to be evaluated over.

   double** unique_data_points;//matrix holding all unique data points from the input matrix.

   int* num_unique_data_points;//jth entry gives the number of data points which are unique in their jth component.

   int predictions_mx_rows, predictions_mx_cols, predictions_update_cols, prediction_fn_value_col;//dimensions of predictions_mx, the number of columns at the end which should be updated each iteration, and the index of the one representing the estimated regression function.
   
   long replace_with_bootstrap_samples;

   long thin_data;// if > 0, only keep those data points for each subject whose row number is a multiple of thin_data (including the zeroth row).

   static long test_grid_size;

   int num_test_points;//num_test_points = test_grid_size or test_grid_size^2.

   long variable_to_predict_over;//when predicting over grid, range over these variables.
   long second_variable_to_predict_over;

   double current_regression_fn_val;//double holding the most recent evaluation of the regression function.

   int regression_num_data_blocks;

   int regression_num_subjects;

   long subject_grouping;//every subject_grouping observations assumed from the same subject.

   int num_subjects;// = num_data_blocks, unless we have read in some Observation_index vector, in which case num_observations is size of that vector.

   int* Subject_index;//int array holding the data row number where each subject begins.

   int* Subject_index_original;//Copy of original Subject_index array if missing values are then deleted.

   int* experiment_numbers;// jth entry gives the experiment number which the jth obervation belongs to.

   int* experiment_numbers_original;// Copy of original experiment_numbers array if missing values are then deleted.

   long num_experiments, num_experiments_original;//number of observations per data block.

   bool treat_experiments_separately;//if non zero, treat each observation per subject as being from a different distribution (1st obsns are distributed one way, 2nd another, and so on..)

   bool correlate_experiments, correlate_columns;//if non zero, specify a correlation between the experiments, columns.

   bool weight_experiments;// if non-zero, weight the observations from each experiment.

   bool standardize_experiments;//if true, make the s.d.'s of the targets for each experiment the same.

   bool treat_experiments_as_different_dimensions;//if true, consider multiple experiments as observations along the axes of different dimensions of a multivariate surface, and treat as just one "experiment".

   bool seemingly_unrelated_regressions;
   
   bool intercepts_for_each_dimension;//if true, then the regression on each x dimension has it's own intercept term, only useful for treat_experiments_as_different_dimensions.

   bool linear_term_for_each_dimension;//if true, then the regression on each x dimension has it's own linear slope term, only useful for treat_experiments_as_different_dimensions.

   int active_dimension;// used if intercepts_for_each_dimension.

   int** subject_exp_num_missing_values;//a vector with ith entry the number of missing values for the ith subject.

   bool random_effects_model;//if true, assume each subject has a random effects parameter.

   double random_effects_precision;//the precision of the random effects coefficients.

   double* random_effect_variances, *random_effect_log_variances;//alternatively to random_effects_precision, a vector holding individual-specific random effect variances.

   double* random_effect_posterior_precisions;//1 + 1/random effect variances.

   double sum_log_random_effect_variances;

   int* equivalent_experiments;//if activated, then experiment i is grouped together with *equivalent_experiments[i].

   double* experiment_weights;// double array containing weights for each experiment, 1/the sd of that experiment.

   double* experiment_sds;//double array containing the original sds for each experiment.

   double log_ave_experiments_sds;//log of average of the experiment sds.

   double* experiment_means;//double array containing the original means for each experiment.

   int current_experiment;//useful global variable to denote the experiment currently under consideration.
   
   long experiment_location_dimension;//if >0, create abstract spatial locations for each experiment in this dimensional space, with closeness in locations of experiments indicating correlation between them.

   long num_locations;//the number of abstract locations, equal to num_experiments if > 1, otherwise p_y.

   long num_location_parameters;//the number of parameters controlling the abstaract experiment locations.

   int* experiment_partition;//a partition of the experiments, used for allocting experiment location means to experiemnt_location_parameters rows.

   int experiment_partition_size;//the number of sets in the experiment_partition.

   int* experiment_partition_sizes;//the sizes of each set in the experiment_partition.

   double* experiment_closeness_power;//power of distance between locations used for calculating experiment correlation -if NULL, power assumed to be 1 in calculations.

   double **experiment_locations, **best_experiment_locations;// matrix for storing abstract spatial locations of experiments.

   double** experiment_location_means;// matrix for storing population means of abstract spatial locations of experiments.

   double** experiment_location_sums;// matrix for storing sample sums of abstract spatial locations of experiments for each group in a partition of the experiments, or the overall sample sum if there is no such partition.

   double** experiment_location_parameters;//matrix for parameters controlling means of abstract spatial locations of experiments.

   int* experiment_signs;//takes value +-1 for each experiment, to define positive or negative correlation.

   double* experiment_var_multipliers, *best_experiment_var_multipliers;// if using kriging model for experiment correlations, these hold the variances of each experiment.
   
   gsl_matrix *experiment_covariances, *mean_experiment_covariances, *best_experiment_covariances, *experiment_precisions;

   bool kriging_experiment_correlation;//if true, use kriging model to specify experiment correlations.

   bool start_experiments_independent;//if true, assume that experiments have initially no correlation.
   
   double experiment_location_precision;//prior precision for experiemnt location values.

   double experiment_parameter_precision;//prior precision for experiemnt location paramter values.

   bool normalise_data_blocks;

   long log_targets, log_inputs;//if equal to 1, do log(x); 2, do log(log(x)); 3, do log(1+x)

   bool logistic_transform_targets, logistic_transform_predictors;//if true, do y=(1+exp(-y))^-1.
   
   bool use_differences, subtract_intercept;

   double **y_differenced_points, **x_differenced_points;//if use_differences true, then put the subtracted out values of y and x in these arrays.

   double **y_block_norms;

   int num_differenced_points_stored;//the number of rows in the above arrays.

   bool predict_over_grid;

   bool predict_on_data;//If true, calculate predicted target values for each of the input data points.

   bool gram_schmidt;// if true, use the Gram-Schmidt process to orthogonalize the design matrix.  

   bool calculate_pairwise_variances;//if true, calculate the sample variance for every pair of subjects and write to file.

   double LQ_pairwise_variance, median_pairwise_variance;//the lower quartile and median subject pairwise variances.

   char ParameterFilename[FILENAMELENGTH];
   ///
   char Subject_Index_Filename[FILENAMELENGTH];//file for reading in subject indices.

   char Experiment_Number_Filename[FILENAMELENGTH];//file for reading in experiment_num vector.

   char Equivalent_Experiments_Filename[FILENAMELENGTH];//file for reading in the integer array of equivalent_experiments.

   char Experiment_Locations_Filename[FILENAMELENGTH];//file for outputting principal components of experiment locations.

   char Experiment_Locations_Input_Filename[FILENAMELENGTH];//file for reading in experiment locations.
   
   char Experiment_Parameters_Filename[FILENAMELENGTH];//file for outputting principal components of experiment location parameters.

   char Experiment_Parameters_Input_Filename[FILENAMELENGTH];//file for reading in experiment location parameters.
   
   char Experiment_Signs_Filename[FILENAMELENGTH];//file for taking signs assigned to experiments for specifying correlations.
   
   char Experiment_Correlations_Filename [FILENAMELENGTH];//file for taking 'correlation matrix' of experiments.

   char Experiment_Partition_Filename [FILENAMELENGTH];//file for reading in a parition of the experiments.
      
   char Residuals_Filename[FILENAMELENGTH];//file for writing residuals matrix.
   
   char Variogram_Filename[FILENAMELENGTH];

   char Pairwise_Variances_Filename[FILENAMELENGTH];

   char Random_Effect_Variances_Filename[FILENAMELENGTH];//file containing individual-specific random effect precisions.

   ///
};   

gsl_matrix* matrix_join( gsl_matrix*& m1, gsl_matrix* m2, bool side = false, bool new_matrix = false );
void matrix_delete_column( gsl_matrix*& m, int col );//deletes column col from the matrix m.
gsl_matrix* matrix_block( gsl_matrix*& m, int n_times, bool overwrite = true );//create n_times copies of m in a block diagonal matrix.
gsl_matrix* matrix_block( gsl_matrix*& m1, gsl_matrix*& m2, bool overwrite_first = true, bool delete_second = true );//create block diagonal matrix of m1 and m2.
gsl_matrix* matrix_block( list<gsl_matrix*>& m_list, bool matrices_same_size );//create block diagonal matrix from the matrices in m_list.
void get_rows( gsl_matrix* m_src, int row_start, int row_end, gsl_matrix*& m_tgt );//creates the submatrix of m_src defined by the row range, returns it in m_tgt. */
void get_submatrix( gsl_matrix* m_src, bool* row_included, bool* col_included, gsl_matrix*& m_tgt, bool negate_row = false, bool negate_col = false );
    
void draw_histogram( double* v, int legnth_v );

void my_invert( gsl_matrix m, gsl_matrix* m_inv );

void my_invert( gsl_matrix* m, gsl_matrix* m_inv );

double my_lndet( gsl_matrix* m );

bool diagonalise( gsl_matrix* m, double threshold = -1 );//zeros all the off diagonal terms (subject to the sbsolute values being less than the threshold if this parameter is used); returns true if the diagonalisation went ahead.

void square_matrix_stripe( gsl_matrix* m, int offset, double val );//sets the offset diagonals' terms to val, assumes square matrix.

void matrix_square_symmetric_strip( gsl_matrix* m, int row, int col, int length, double val );//sets diagonal strips of length, taking value val, starting from (row,col) and (col,row).

bool gsl_matrices_differ( const gsl_matrix* m1, const gsl_matrix* m2 );//returns true iff m1 and m2 are identical.

bool matrix_cols_equal_up_to_pairs( const gsl_matrix* m, int col1, int col2 );// returns true iff the two cols of m are equal.

bool matrix_cols_almost_equal( const gsl_matrix* m, int col1, int col2, double tolerance );//returns true if the two cols of m have at least 1-tolerance proportion of their entries equal.

void gram_schmidt_process( gsl_matrix* M, double** coeff_matrix, bool orthonormalise = true, unsigned int gram_rows = 0, unsigned int num_blocks = 1, unsigned int* blocks = NULL, int orthonormlised_blocks = 1 );//performs the Gram Schmidt orthogonalisation process on the columns of M. The coefficients of the new bases corresponding to the columns of M are put in coeff_matrix, where the i'th row corresponds to the ith new base. If blocks is used, this informs a block diagonal structure of M to ease computation. If orthonormlised_blocks is used, this forces the routine to imagine M is roughly made up of this many copies of a matrix, and so to orthonormalise for one such block.

double log_transform( double x, int log_transform_type );//offers different versions of a log transfrom.

double inverse_log_transform( double x, int log_transform_type );//inverts the above function.

double logistic_transform( double x );

double logit_transform( double x );

double sum_gsl_matrix( gsl_matrix*& m );

template <class T> void bounce( T& val, const T lower, const T upper )
{
   bool less = false, greater = false;
   if( val < lower )
      less = true;
   else if( val > upper )
      greater = true;
   while( less || greater )
   {
      if( less )
      {
         val = 2 * lower - val;
         less = false;
         greater = val > upper;
      }
      else
      {
         val = 2 * upper - val;
         greater = false;
         less = val < lower;
      }
   }
}

template <class T> void create_array( T**& T_array, const int dim1, const int dim2 = 0 )
{
   T_array = new T*[ dim1 ];
   if( dim2 > 0 )
      for( int i = 0; i < dim1; i++ )
         T_array[ i ] = new T[ dim2 ];
}

template <class T> void create_ragged_array( T**& T_array, const int dim1, int* dim2 = NULL )
{
   T_array = new T*[ dim1 ];
   if( dim2 )
      for( int i = 0; i < dim1; i++ )
         T_array[ i ] = new T[ dim2[ i ] ];
}

template <class T> void create_zeroed_array( T*& T_array, const int dim1 )
{
   T_array = new T[ dim1 ];
   for( int i = 0; i < dim1; i++ )
      T_array[ i ] = 0;
}

template <class T> void create_null_array( T*& T_array, const int dim1 )
{
   T_array = new T[ dim1 ];
   for( int i = 0; i < dim1; i++ )
      T_array[ i ] = NULL;
}

template <class T> void create_indexing_array( T*& T_array, const int dim1 )
{
   T_array = new T[ dim1 ];
   for( int i = 0; i < dim1; i++ )
      T_array[ i ] = i;
}

template <class T> void create_zeroed_array( T**& T_array, const int dim1, const int dim2 = 0 )
{
   T_array = new T*[ dim1 ];
   for( int i = 0; i < dim1; i++ )
   {
      if( dim2 > 0 )
      {
         T_array[ i ] = new T[ dim2 ];
         for( int j = 0; j < dim2; j++ )
            T_array[ i ][ j ] = 0;
      }
      else
         T_array[ i ] = 0;
   }
}

template <class T> void zero_array( T*& T_array, const int dim1 )
{
   for( int i = 0; i < dim1; i++ )
      T_array[ i ] = 0;
}

template <class T> void zero_array( T**& T_array, const int dim1, const int dim2)
{
   for( int i = 0; i < dim1; i++ )
      for( int j = 0; j < dim2; j++ )
         T_array[ i ][ j ] = 0;
}

template <class T> void scale_array( T*& T_array, T d, const int dim1 )
{
   for( int i = 0; i < dim1; i++ )
      T_array[ i ] *= d;
}

template <class T> void scale_array( T**& T_array, T d, const int dim1, const int dim2 )
{
   for( int i = 0; i < dim1; i++ )
      for( int j = 0; j < dim2; j++ )
	 T_array[ i ][ j ] *= d;
}

template <class T> void add_array( T*& T_array1, T*& T_array2, const int dim1 )
{
   for( int i = 0; i < dim1; i++ )
      T_array1[ i ] += T_array2[ i ];
}

template <class T> void add_array( T**& T_array1, T**& T_array2, const int dim1, const int dim2 )
{
   for( int i = 0; i < dim1; i++ )
      for( int j = 0; j < dim2; j++ )
	 T_array1[ i ][ j ] += T_array2[ i ][ j ];
}

template <class T> void add_scaled_array( T*& T_array1, T*& T_array2, T t, const int dim1 )
{
   for( int i = 0; i < dim1; i++ )
      T_array1[ i ] += t * T_array2[ i ];
}

template <class T> void subtract_array( T*& T_array1, T*& T_array2, const int dim1 )
{
   for( int i = 0; i < dim1; i++ )
      T_array1[ i ] -= T_array2[ i ];
}

template <class T> void subtract_array( T**& T_array1, T**& T_array2, const int dim1, const int dim2 )
{
   for( int i = 0; i < dim1; i++ )
      for( int j = 0; j < dim2; j++ )
	 T_array1[ i ][ j ] -= T_array2[ i ][ j ];
}

template <class T> void subtract_scaled_array( T*& T_array1, T*& T_array2, T t, const int dim1 )
{
   for( int i = 0; i < dim1; i++ )
      T_array1[ i ] -= t * T_array2[ i ];
}

template <class T> T sum_array( T*& T_array, const int dim1 )
{
   T t = 0;
   for( int i = 0; i < dim1; i++ )
      t += T_array[ i ];
   return t;
}

template <class T> T sum_array( T**& T_array, const int dim1, const int dim2  )
{
   T t = 0;
   for( int i = 0; i < dim1; i++ )
      for( int j = 0; j < dim2; j++ )
         t += T_array[ i ][ j ];
   return t;
}

template <class T> void zap_array( T**& T_array, const int dim1, const int dim2 = 0 )
{
   if( !T_array )
      return;
   if( dim2 > 0 )
   {
      for( int i = 0; i < dim1; i++ )
         if( T_array[ i ] )
            delete [] T_array[ i ];
   }
   delete [] T_array;
   T_array = 0;
}

template <class T> void create_array( T***& T_array, const int dim1, const int dim2, const int dim3 )
{
   T_array = new T**[ dim1 ];
   if( dim2 > 0 )
      for( int i = 0; i < dim1; i++ )
      {
         T_array[ i ] = new T*[ dim2 ];
         if( dim3 > 0 )
            for( int j = 0; j < dim2; j++ )
               T_array[ i ][ j ] = new T[ dim3 ];
      }
}

template <class T> void create_zeroed_array( T***& T_array, const int dim1, const int dim2, const int dim3 )
{
   T_array = new T**[ dim1 ];
   for( int i = 0; i < dim1; i++ )
   {
      if( dim2 > 0 )
      {
         T_array[ i ] = new T*[ dim2 ];
         for( int j = 0; j < dim2; j++ )
            if( dim3 > 0 )
            {
               T_array[ i ][ j ] = new T[ dim3 ];
               for( int k = 0; k < dim3; k++ )
                  T_array[ i ][ j ][ k ] = 0;
            }
            else
               T_array[ i ][ j ] = 0;
      }
      else
         T_array[ i ] = 0;
   }
}

template <class T> void zero_array( T***& T_array, const int dim1, const int dim2, const int dim3 )
{
   for( int i = 0; i < dim1; i++ )
      for( int j = 0; j < dim2; j++ )
         for( int k = 0; k < dim3; k++ )
            T_array[ i ][ j ][ k ] = 0;
}

template <class T> void scale_array( T***& T_array, T d, const int dim1, const int dim2, const int dim3 )
{
   for( int i = 0; i < dim1; i++ )
      for( int j = 0; j < dim2; j++ )
         for( int k = 0; k < dim3; k++ )
            T_array[ i ][ j ][ k ] *= d;
}

template <class T> void zap_array( T***& T_array, const int dim1, const int dim2, const int dim3 )
{
   if( !T_array )
      return;
   if( dim2 > 0 )
   {
      for( int i = 0; i < dim1; i++ )
      {
         if( dim3 > 0 )
         {
            for( int j = 0; j < dim2; j++ )
               if( T_array[ i ][ j ] )
                  delete [] T_array[ i ][ j ];
         }
         if( T_array[ i ] )
            delete [] T_array[ i ];
      }
   }
   delete [] T_array;
   T_array = 0;
}

template <class T> void print_array( T*& T_array, const int dim1 )
{
   if( !T_array )
   {
       cout << "Null" << endl;
       return;
   }
   for( int i = 0; i < dim1; i++ )
      cout << T_array[ i ] << "  ";
   cout << endl;
}

template <class T> void print_array( T**& T_array, const int dim1, const int dim2 )
{
   if( !T_array )
   {
       cout << "Null" << endl;
       return;
   }
   for( int i = 0; i < dim1; i++ )
   {
      for( int j = 0; j < dim2; j++ )
         cout << T_array[ i ][ j ] << "  ";
      cout << endl;
   }
}

template <class T> void print_array( T***& T_array, const int dim1, const int dim2, const int dim3 )
{
   if( !T_array )
   {
       cout << "Null" << endl;
       return;
   }
   for( int i = 0; i < dim1; i++ )
   {
      for( int j = 0; j < dim2; j++ )
      {
         for( int k = 0; k < dim3; k++ )
            cout << T_array[ i ][ j ][ k ] << "  ";
         cout << endl;
      }
      cout << endl;
      cout << endl;
   }
}

template <class T> void print_triangular_array( T**& T_array, const int dim )
{
   if( !T_array )
   {
       cout << "Null" << endl;
       return;
   }
   for( int i = 0; i < dim; i++ )
   {
      if( i==0 || T_array[ i ] )
      {
         for( int j = 0; j < dim; j++ )
            if( j == 0 || T_array[ j ] )
            {
               if( i > j )
                  cout << T_array[ i ][ j ] << "  ";
               else if( i < j )
                  cout << T_array[ j ][ i ] << "  ";
//               else
//                  cout << 0 << "  ";
            }
         cout << endl;
      }
   }
}

template <class T> void copy_array( T*& T_array_copy, T* T_array, const int dim1 )
{
   if( !T_array_copy )
      T_array_copy = new T[ dim1 ];
   for( int i = 0; i < dim1; i++ )
      T_array_copy[ i ] = T_array[ i ];
}

template <class T> void copy_array( T**& T_array_copy, T** T_array, const int dim1, const int dim2 )
{
   if( !T_array_copy )
      create_array( T_array_copy, dim1, dim2 );
   for( int i = 0; i < dim1; i++ )
      for( int j = 0; j < dim2; j++ )
         T_array_copy[ i ][ j ] = T_array[ i ][ j ];
}

template <class T> void copy_array( T***& T_array_copy, T*** T_array, const int dim1, const int dim2, const int dim3 )
{
   if( !T_array_copy )
      create_array( T_array_copy, dim1, dim2, dim3 );
   for( int i = 0; i < dim1; i++ )
      for( int j = 0; j < dim2; j++ )
          for( int k = 0; k < dim3; k++ )
             T_array_copy[ i ][ j ][ k ] = T_array[ i ][ j ][ k ];
}

template <class T> void create_triangular_array( T**& T_array, const int dim1 )
{
   T_array = new T**[ dim1 ];
   T_array[ 0 ] = 0;
   for( int i = 1; i < dim1; i++ )
      T_array[ i ] = new T[ i ];
}

template <class T> void copy_triangular_array( T**& T_array_copy, T** T_array, const int dim1 )
{
   bool create_flag = false;
   if( !T_array_copy )
   {
      create_flag = true;
      T_array_copy = new T*[ dim1 ];
      T_array_copy[ 0 ] = 0;
   }
   for( int i = 1; i < dim1; i++ )
   {
      if( create_flag || !T_array_copy[ i ] )
         T_array_copy[ i ] = new T[ i ];
      for( int j = 0; j < i; j++ )
         T_array_copy[ i ][ j ] = T_array[ i ][ j ];
   }
}

template <class T> void zap_triangular_array( T**& T_array, const int dim1 )
{
   if( !T_array )
      return;
   for( int i = 1; i < dim1; i++ )
      if( T_array[ i ] )
         delete [] T_array[ i ];
   delete [] T_array;
   T_array = 0;
}

template <class T> void copy_scaled_array( T*& T_array_copy, T* T_array, T d, const int dim1 )
{
   if( !T_array_copy )
      T_array_copy = new T[ dim1 ];
   for( int i = 0; i < dim1; i++ )
         T_array_copy[ i ] = T_array[ i ] * d;
}

template <class T> void copy_scaled_array( T**& T_array_copy, T** T_array, T d, const int dim1, const int dim2 )
{
   if( !T_array_copy )
      create_array( T_array_copy, dim1, dim2 );
   for( int i = 0; i < dim1; i++ )
      for( int j = 0; j < dim2; j++ )
         T_array_copy[ i ][ j ] = T_array[ i ][ j ] * d;
}

template <class T> void copy_scaled_array( T***& T_array_copy, T*** T_array, T d, const int dim1, const int dim2, const int dim3 )
{
   if( !T_array_copy )
      create_array( T_array_copy, dim1, dim2, dim3 );
   for( int i = 0; i < dim1; i++ )
      for( int j = 0; j < dim2; j++ )
          for( int k = 0; k < dim3; k++ )
             T_array_copy[ i ][ j ][ k ] = T_array[ i ][ j ][ k ] * d;
}

template <class T> T get_array_norm( T*& T_array, const int dim1 )
{
   T sum = 0;
   for( int i = 0; i < dim1; i++ )
      sum += T_array[ i ] * T_array[ i ];
   return sum;
}
 
template <class T> T get_array_difference_norm( T*& T_array1, T*& T_array2, const int dim1 )
{
   T sum = 0;
   for( int i = 0; i < dim1; i++ )
   {
      T diff = T_array1[ i ] - T_array2[ i ];
      sum += diff * diff;
   }
   return sum;
}

template <class T> void array_delete_column( T**& m, int col, int rows, int cols, bool soft = false )//removes the column col from the array - if soft = true, keep the matrix dimension the same and fill the last column with zeros.
{
   if( !m )
      return;
   int m_new_rows = rows;
   int m_new_cols = cols - 1;
   if( m_new_cols < 1 || col >= m_new_cols + 1 )
   {
      cout << "Can't delete column " << col << " , matrix doesn't have that many columns"<< endl;
      return;
   }
   T** mcopy = m;
   if( !soft )
      create_array( m, m_new_rows, m_new_cols );
   int j;
   for( int i = 0; i < m_new_rows; i++ )
   {
      if( !soft )
         for( j = 0; j < col; j++ )
             m[i][j] = mcopy[i][j];
      for( j = col; j < m_new_cols; j++ )
          m[i][j] = mcopy[i][j+1];
      if( soft )
          m[i][m_new_cols] = 0;
   }
   if( !soft )
      zap_array( mcopy, rows, cols );
}

template <class T> void array_insert_column( T**& m, unsigned int col, unsigned int rows,  unsigned int cols, bool soft = false )//adds a column at position col to the array - if soft = true, keep the matrix dimension the same, assuming it is already big enough.
{
   if( !m )
      return;
   unsigned int m_new_rows = rows;
   unsigned int m_new_cols = cols + 1;
   if( col > cols )
   {
      cout << "Can't instert column at position " << col << " , matrix doesn't have that many columns"<< endl;
      return;
   }
   T** mcopy = m;
   if( !soft )
      create_array( m, m_new_rows, m_new_cols );
   unsigned int j;
   for( unsigned int i = 0; i < m_new_rows; i++ )
   {
      if( !soft )
         for( j = 0; j < col; j++ )
             m[i][j] = mcopy[i][j];
      for( j = m_new_cols - 1; j > col; j-- )
          m[i][j] = mcopy[i][j-1];
      m[i][col] = 0;
   }
   if( !soft )
      zap_array( mcopy, rows, cols );
}

template <class T> void array_delete_row( T**& m, int row, int rows )
{
   if( !m )
      return;
   if( rows <= 1 || row >= rows )
   {
      cout << "Can't delete row " << row << " , matrix doesn't have that many rows"<< endl;
      return;
   }
   delete [] m[row];
   for( int i = row; i < (rows-1); i++ )
      m[ i ] = m[ i+1 ];
}


template <class T> void array_delete_row( T**& m, int row, int rows, int cols, bool soft = false )//removes the row 'row' from the array - if soft = true, keep the matrix dimension the same and fill the last row with zeros.
{
   if( !m )
      return;
   int m_new_rows = rows - 1;
   int m_new_cols = cols;
   if( rows <= 1 || row >= rows )
   {
      cout << "Can't delete row " << row << " , matrix doesn't have that many rows"<< endl;
      return;
   }

   T* row_ptr = NULL;

   if( !soft )
      delete [] m[row];
   else
      row_ptr = m[row];
   //      create_array( m, m_new_rows, m_new_cols );
   for( int i = row; i < m_new_rows; i++ )
      m[ i ] = m[ i+1 ];
   m[ m_new_rows ] = row_ptr;
   return;
   
   int i;
   T** mcopy = m;
   for( int j = 0; j < m_new_cols; j++ )
   {
      if( !soft )
         for( i = 0; i < row; i++ )
             m[i][j] = mcopy[i][j];
      for( i = row; i < m_new_rows; i++ )
          m[i][j] = mcopy[i+1][j];
      if( soft )
          m[m_new_rows][j] = 0;
   }
   if( !soft )
      zap_array( mcopy, rows, cols );
}

template <class T> void array_insert_row( T**& m, unsigned int row, unsigned int rows,  unsigned int cols, bool soft = false )//adds a row at position row to the array - if soft = true, keep the matrix dimension the same, assuming it is already big enough.
{
   if( !m )
      return;
   unsigned int m_new_rows = rows + 1;
   unsigned int m_new_cols = cols;
   if( row > rows )
   {
      cout << "Can't instert row at position " << row << " , matrix doesn't have that many rows"<< endl;
      return;
   }
   T** mcopy = m;
   if( !soft )
      create_array( m, m_new_rows, m_new_cols );
   unsigned int i;
   for( unsigned int j = 0; j < m_new_cols; j++ )
   {
      if( !soft )
         for( i = 0; i < row; i++ )
             m[i][j] = mcopy[i][j];
      for( i = m_new_rows - 1; i > row; i-- )
          m[i][j] = mcopy[i-1][j];
      m[row][j] = 0;
   }
   if( !soft )
      zap_array( mcopy, rows, cols );
}



#endif
